<?php
// veriword.lib.php
// Verification word generation class
//
// This class was taken from http://www.program-ruti.org/veriword/
// and modified to properly fit UNB's environment.

## v1.2 Stable ## 18 August 2004
## v1.1 Stable ## 01 August 2004
## v1	Stable ## 27 July 2004
######
## Verification Word
######
## This class generate an image with random text
## to be used in form verification. It has visual
## elements design to confuse OCR software preventing
## the use of BOTS.
##
#######
## Author: Huda M Elmatsani
## Email: 	justhuda at netrada.co.id
##
## 25/07/2004
#######
## Copyright (c) 2004 Huda M Elmatsani All rights reserved.
## This program is free for any purpose use.
########
##
## USAGE
## create some image with noise texture, put in image directory,
## put some true type font into font directory,
## you can search and put free font you like
##
##
## see sample.php for test and usage
##
####

class VeriWord
{
	var $dir_font;   // path to font directory, supports *.ttf
	var $dir_noise;   // path to background image directory, supports *.jpg|jpeg|png

	var $word = '';
	var $im_width = 0;
	var $im_height = 0;
	var $im_type = '';   //image type: jpeg, png

	// Added [YG]
	var $initial_textsize = 25;   // bigger value makes smaller text
	var $text_alpha = 85;         // blending % of text over the noise
	var $vert_offset = 5;         // text vertical offset
	var $def_width = 70;         // default image width
	var $def_height = 25;         // default image height

	function VeriWord($w = 0, $h = 0)
	{
		$this->dir_font = dirname(__FILE__) . '/fonts/';
		$this->dir_noise = dirname(__FILE__) . '/noises/';

		if (!$w) $w = $this->def_width;
		if (!$h) $h = $this->def_height;

		// create session to set word for verification
		$prog_id = $_GET['prog_id'];
		session_name($prog_id . 'sess');
		session_start();

		$this->word = $this->pick_word();
		$_SESSION['veriword'] = $this->word;

		$this->im_width = $w;
		$this->im_height = $h;
	}

	function output_image($type = 'jpeg')
	{
		// make it not case sensitive
		$this->im_type = strtolower($type);

		// check image type availability
		$this->validate_type();

		// draw the image
		$this->draw_image();

		@header('Pragma: no-cache');
		@header('Cache-Control: no-cache');
		@header('Expires: ' . date('r'));   // that is now

		// show the image
		switch ($this->im_type)
		{
		case 'png':
			header('Content-type: image/png');
			imagepng($this->im);
			break;
		case 'jpeg':
		case 'jpg':
		default:
			header('Content-type: image/jpeg');
			imagejpeg($this->im);
			break;
		}
	}

	function pick_word()
	{
		$symbols = array('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'C', 'E', 'F', 'H', 'J', 'K', 'L', 'M', 'N', 'P', 'R', 'T', 'U', 'X', 'Y');

		$word = '';
		while (strlen($word) < 4)
		{
			$ch = rand(0, sizeof($symbols) - 1);
			$word .= $symbols[$ch];
		}
		return $word;

		#return rand(1000,9999);
	}

	// pick one noise image randomly from image directory
	// note: please avoid file name space
	//
	function get_font()
	{
		$exts = array('ttf', 'TTF');
		$filelist = $this->scandir($this->dir_font, $exts);
		if (sizeof($filelist) > 0)
		{
			srand((double) microtime() * 1000000);
			return $this->dir_font . $filelist[array_rand($filelist)];
		}
		else
			return 5;
	}

	function draw_text()
	{
		// pick one font type randomly from font directory
		$text_font 	= $this->get_font();
		// angle for text inclination
		$text_angle = rand(-4,4);
		$this->vert_offset += $text_angle / 2;
		if (!$text_angle) $text_angle = 1;

		// create canvas for text drawing
		$im_text = @imagecreate($this->im_width, $this->im_height);
   		$bg_color = imagecolorallocate($im_text, 255, 255, 255);

		// pick color for text
		$text_color = imagecolorallocate($im_text, 0, 0, 0);

		// numeric means built-in font
		if (is_numeric($text_font))
		{
			$text_width = imagefontwidth($text_font) * strlen($this->word);
			$text_height = imagefontheight($text_font);
			$margin = $text_width * 0.2;
			$im_string = @imagecreatetruecolor($text_width + $margin, $text_height + $margin);

			// calculate center position of text
			$text_x = $margin/2;
			$text_y = $margin/2;

			imagestring($im_string, $text_font, $text_x, $text_y, $this->word, $text_color);

			imagecolortransparent($im_string, $bg_color);

			// resize the text, because built-in font is very small
			imagecopyresampled(
				$im_text,
				$im_string,
				0, 0, 0, 0,
				$this->im_width,
				$this->im_height,
				$text_width+$margin,
				$text_height+$margin);
		}
		else
		{
			// initial text size
			$text_size = $this->initial_textsize;
			// calculate text width and height
			$box = imagettfbbox ( $text_size, $text_angle, $text_font, $this->word);
			$text_width = $box[2] - $box[0];   // text width
			$text_height = $box[5] - $box[3];   // text height

			// adjust text size
			$text_size = round((20 * $this->im_width) / $text_width) * (rand(95, 105) / 100);

			// recalculate text width and height
			$box = imagettfbbox ( $text_size, $text_angle, $text_font, $this->word);
			$text_width = $box[2] - $box[0];   // text width
			$text_height = $box[5] - $box[3];   // text height

			// calculate center position of text
			$text_x = ($this->im_width - $text_width) / 2 + rand(-2,8);
			$text_y = ($this->im_height - $text_height) / 2 + rand(-2,2) + $this->vert_offset;   // added vertical offset [YG]

			// draw text into canvas
			imagettftext(
				$im_text,
				$text_size,
				$text_angle,
				$text_x,
				$text_y,
				$text_color,
				$text_font,
				$this->word);
		}

		// draw angled lines with random degree and various y position)
		/*
		imagesetthickness($im_text, 1);
		$line_color = imagecolorallocate($im_text, 128, 128, 128);
		imageline($im_text, 0, $this->im_height / 2 + rand(-15, 15), $this->im_width, $this->im_height / 2 + rand(-5, 5), $line_color);
		imageline($im_text, 0, $this->im_height / 2 + rand(-15, 15), $this->im_width, $this->im_height / 2 + rand(-5, 5), $line_color);
		*/

		// remove background color
		imagecolortransparent($im_text, $bg_color);
		return $im_text;
		#imagedestroy($im_text);   // what's this??
	}

	// pick one noise image randomly from image directory
	// Suggested by Gadi Cohen <dragon at wastelands dot net>
	//
	function get_noise()
	{
		$exts = array('jpg$|\\.jpeg', 'png');
	   	$filelist = $this->scandir($this->dir_noise, $exts);
	   	if (sizeof($filelist) > 0)
	   	{
			srand((double) microtime() * 1000000);
			return $this->dir_noise . $filelist[array_rand($filelist)];
		}
	   	else
	   		return false;
	}

	function draw_image()
	{
		// get the noise image file
		$img_file = $this->get_noise();

		if ($img_file)
		{
			$extension = $this->get_imgext($img_file);

			// create "noise" background image from your image stock
			switch ($extension)
			{
			case 'jpeg':
			case 'jpg':
				$im_noise = @imagecreatefromjpeg ($img_file);
				break;
			case 'png':
				$im_noise = @imagecreatefrompng ($img_file);
				break;
			}
		}
		else
		{
			// if fail to load image file, create it on the fly
			$im_noise = $this->draw_noise();
		}

 		$noise_width = imagesx($im_noise);
		$noise_height = imagesy($im_noise);

		// resize the background image to fit the size of image output
		$this->im = @imagecreatetruecolor($this->im_width,$this->im_height);
		imagecopyresampled(
			$this->im,
			$im_noise,
			0, 0, 0, 0,
			$this->im_width,
			$this->im_height,
			$noise_width,
			$noise_height);

		// put text image into background image
		imagecopymerge(
			$this->im,
			$this->draw_text(),
			0, 0, 0, 0,
			$this->im_width,
			$this->im_height,
			$this->text_alpha);

		return $this->im;
		#imagedestroy($this->im);   // what's this??
	}

	// create "noise" background image
	//
	function draw_noise()
	{
		$im_noise = @imagecreatetruecolor($this->im_width, $this->im_height);
		$bg_color = imagecolorallocate($im_noise, 255, 255, 255);
		imagefill($im_noise, 0, 0, $bg_color);

		for ($i = 0; $i < $this->im_height; $i++)
		{
			$c = rand(140, 220);
			$line_color = imagecolorallocate($im_noise, $c, $c, $c);
			imagesetthickness($im_noise, rand(1, 5));
			imageline(
				$im_noise,
				rand(0, 30),                     // x1
				$i + rand(-10, 10),              // y1
				$this->im_width - rand(0, 30),   // x2
				$i + rand(-10, 10),              // y2
				$line_color);                    // color
		}
		for ($i = 0; $i < 1000; $i++)
		{
			$c = rand(160, 220);
			$line_color = imagecolorallocate($im_noise, $c, $c, $c);
			imagesetpixel(
				$im_noise,
				rand(0, $this->im_width),    // x
				rand(0, $this->im_height),   // y
				$line_color);                // color
		}

		return $im_noise;
		#imagedestroy($im_noise);   // what's this??
	}

	// check image type availability
	//
	function validate_type()
	{
		$is_available = false;

		switch ($this->im_type)
		{
			case 'jpeg' :
			case 'jpg' 	:
				if(function_exists('imagejpeg'))
				$is_available = true;
				break;
			case 'png' :
				if(function_exists('imagepng'))
				$is_available = true;
				break;
		}
		if (!$is_available && function_exists('imagejpeg'))
		{
			// if not available, cast image type to jpeg
			$this->im_type = 'jpeg';
			return true;
		}
		else if (!$is_available && function_exists('imagepng'))
		{
			// if not available, cast image type to png
			$this->im_type = 'png';
			return true;
		}
		else if(!$is_available && !function_exists('imagejpeg'))
		{
		   die('No image support in this PHP server');
		}
		else
			return true;
	}

	// function of rolandfx_SPAMMENOT_ at bellsouth dot net
	//
	function scandir($adir, $exts)
	{
		$dir = @opendir($adir);
		while (($files = @readdir($dir)) !== false)
		{
			foreach ($exts as $value)
			{
				if ($this->check_ext($files, $value))
				{
					$match_files[] = $files;
					break;   // No need to keep looping if we've got a match.
				}
			}
		}
		@closedir($dir);
		return $match_files;
	}

	function check_ext($filename, $ext)
	{
		$testext = '\.' . $ext . '$';
		return eregi($testext, $filename);
	}

	function get_imgext($file)
	{
		$acceptable = array('jpg', 'jpeg', 'png');
		// ask the image type
		$file_info = pathinfo($file);
		$extension = $file_info['extension'];

		if (in_array($extension, $acceptable))
			return $extension;
		else
			return null;
	}
}

?>